using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices.ComTypes;
using System.Runtime.InteropServices;
using DirectShowLib;

namespace SoundConv
{
    delegate void WritingGraphPart(IGraphBuilder pGraph, ICaptureGraphBuilder2 pBuilder, string outfile, IBaseFilter pSampleGrabber);

    class OutputFormat
    {
        public WritingGraphPart makeGraphProc;
        string name;

        public OutputFormat(string format_name, WritingGraphPart makeWritingPart)
        {
            makeGraphProc = makeWritingPart;
            name = format_name;
        }

        public override string ToString()
        {
            return name;
        }
    }


    public partial class MainForm : Form
    {
        void WriteWAV(IGraphBuilder pGraph, ICaptureGraphBuilder2 pBuilder, string outfile, IBaseFilter pSampleGrabber)
        {
            int hr = 0;
            //add WAV Dest
            Guid CLSID_WAVDest = new Guid("{3C78B8E2-6C4D-11D1-ADE2-0000F8754B99}"); //wavdest.ax
            IBaseFilter pWAVDest = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_WAVDest));
            hr = pGraph.AddFilter(pWAVDest, "WAV Dest");
            checkHR(hr, "Can't add WAV Dest to graph");

            //connect SampleGrabber and WAV Dest
            hr = pBuilder.RenderStream(null, MediaType.Audio, pSampleGrabber, null, pWAVDest);
            checkHR(hr, "Can't connect SampleGrabber and WAV Dest");

            //add File writer
            IBaseFilter pFilewriter = (IBaseFilter)new FileWriter();
            hr = pGraph.AddFilter(pFilewriter, "File writer");
            checkHR(hr, "Can't add File writer to graph");
            //set destination filename
            IFileSinkFilter pFilewriter_sink = pFilewriter as IFileSinkFilter;
            if (pFilewriter_sink == null)
                checkHR(unchecked((int)0x80004002), "Can't get IFileSinkFilter");
            hr = pFilewriter_sink.SetFileName(outfile, null);
            checkHR(hr, "Can't set filename");

            //connect WAV Dest and File writer
            hr = pBuilder.RenderStream(null, MediaType.Stream, pWAVDest, null, pFilewriter);
            checkHR(hr, "Can't connect WAV Dest and File writer");
        }

        void WriteMP3(IGraphBuilder pGraph, ICaptureGraphBuilder2 pBuilder, string outfile, IBaseFilter pSampleGrabber)
        {
            int hr = 0;
            Guid CLSID_LAMEAudioEncoder = new Guid("{B8D27088-FF5F-4B7C-98DC-0E91A1696286}"); //lame.ax
            Guid CLSID_Dump = new Guid("{36A5F770-FE4C-11CE-A8ED-00AA002FEAB5}"); //dump.ax

            //add LAME Audio Encoder
            IBaseFilter pLAMEAudioEncoder = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_LAMEAudioEncoder));
            hr = pGraph.AddFilter(pLAMEAudioEncoder, "LAME Audio Encoder");
            checkHR(hr, "Can't add LAME Audio Encoder to graph");

            //connect SampleGrabber and LAME Audio Encoder
            hr = pBuilder.RenderStream(null, MediaType.Audio, pSampleGrabber, null, pLAMEAudioEncoder);
            checkHR(hr, "Can't connect SampleGrabber and LAME Audio Encoder");

            //add Dump
            IBaseFilter pDump = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_Dump));
            hr = pGraph.AddFilter(pDump, "Dump");
            checkHR(hr, "Can't add Dump to graph");
            //set destination filename
            IFileSinkFilter pDump_sink = pDump as IFileSinkFilter;
            if (pDump_sink == null)
                checkHR(unchecked((int)0x80004002), "Can't get IFileSinkFilter");
            hr = pDump_sink.SetFileName(outfile, null);
            checkHR(hr, "Can't set filename");

            //connect LAME Audio Encoder and Dump
            hr = pBuilder.RenderStream(null, MediaType.Audio, pLAMEAudioEncoder, null, pDump);
            checkHR(hr, "Can't connect LAME Audio Encoder and Dump");
        }

        void WriteOGG(IGraphBuilder pGraph, ICaptureGraphBuilder2 pBuilder, string outfile, IBaseFilter pSampleGrabber)
        {
            int hr = 0;
            Guid CLSID_VorbisEncodeFilter = new Guid("{5C94FE86-B93B-467F-BFC3-BD6C91416F9B}"); //dsfVorbisEncoder.dll
            Guid CLSID_OggMuxFilter = new Guid("{1F3EFFE4-0E70-47C7-9C48-05EB99E20011}"); //dsfOggMux.dll

            //add Vorbis Encode Filter
            IBaseFilter pVorbisEncodeFilter = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_VorbisEncodeFilter));
            hr = pGraph.AddFilter(pVorbisEncodeFilter, "Vorbis Encode Filter");
            checkHR(hr, "Can't add Vorbis Encode Filter to graph");

            //connect SampleGrabber and Vorbis Encode Filter
            hr = pBuilder.RenderStream(null, MediaType.Audio, pSampleGrabber, null, pVorbisEncodeFilter);
            checkHR(hr, "Can't connect SampleGrabber and Vorbis Encode Filter");

            //add Ogg Mux Filter
            IBaseFilter pOggMuxFilter = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_OggMuxFilter));
            hr = pGraph.AddFilter(pOggMuxFilter, "Ogg Mux Filter");
            checkHR(hr, "Can't add Ogg Mux Filter to graph");
            //set destination filename
            IFileSinkFilter pOggMuxFilter_sink = pOggMuxFilter as IFileSinkFilter;
            if (pOggMuxFilter_sink == null)
                checkHR(unchecked((int)0x80004002), "Can't get IFileSinkFilter");
            hr = pOggMuxFilter_sink.SetFileName(outfile, null);
            checkHR(hr, "Can't set filename");

            //connect Vorbis Encode Filter and Ogg Mux Filter
            hr = pBuilder.RenderStream(null, MediaType.Audio, pVorbisEncodeFilter, null, pOggMuxFilter);
            checkHR(hr, "Can't connect Vorbis Encode Filter and Ogg Mux Filter");
        }

        void WriteAC3(IGraphBuilder pGraph, ICaptureGraphBuilder2 pBuilder, string outfile, IBaseFilter pSampleGrabber)
        {
            int hr = 0;
            Guid CLSID_Dump = new Guid("{36A5F770-FE4C-11CE-A8ED-00AA002FEAB5}"); //dump.ax

            //add AC-3 ACM Codec
            IBaseFilter pAC3ACMCodec = CreateFilter(@"@device:cm:{33D9A761-90C8-11D0-BD43-00A0C911CE86}\8192AC-3 ACM Codec");
            hr = pGraph.AddFilter(pAC3ACMCodec, "AC-3 ACM Codec");
            checkHR(hr, "Can't add AC-3 ACM Codec to graph");

            //connect SampleGrabber and AC-3 ACM Codec
            hr = pBuilder.RenderStream(null, MediaType.Audio, pSampleGrabber, null, pAC3ACMCodec);
            checkHR(hr, "Can't connect SampleGrabber and AC-3 ACM Codec");

            //set format (bitrate)
            AMMediaType pmt = new AMMediaType();
            pmt.majorType = MediaType.Audio;
            pmt.subType = new Guid("{00002000-0000-0010-8000-00AA00389B71}");
            pmt.formatType = FormatType.WaveEx;
            pmt.fixedSizeSamples = true;
            pmt.formatSize = 18;
            pmt.sampleSize = 348;
            pmt.temporalCompression = false;
            WaveFormatEx format = new WaveFormatEx();
            format.wFormatTag = 8192;
            format.nChannels = 2;
            format.nSamplesPerSec = 44100;
            format.nAvgBytesPerSec = 9991;
            format.nBlockAlign = 348;
            pmt.formatPtr = Marshal.AllocCoTaskMem(Marshal.SizeOf(format));
            Marshal.StructureToPtr(format, pmt.formatPtr, false);
            hr = ((IAMStreamConfig)GetPin(pAC3ACMCodec, "Output")).SetFormat(pmt);
            DsUtils.FreeAMMediaType(pmt);
            checkHR(hr, "Can't set format");

            //add Dump
            IBaseFilter pDump = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_Dump));
            hr = pGraph.AddFilter(pDump, "Dump");
            checkHR(hr, "Can't add Dump to graph");
            //set destination filename
            IFileSinkFilter pDump_sink = pDump as IFileSinkFilter;
            if (pDump_sink == null)
                checkHR(unchecked((int)0x80004002), "Can't get IFileSinkFilter");
            hr = pDump_sink.SetFileName(outfile, null);
            checkHR(hr, "Can't set filename");

            //connect AC-3 ACM Codec and Dump
            hr = pBuilder.RenderStream(null, MediaType.Audio, pAC3ACMCodec, null, pDump);
            checkHR(hr, "Can't connect AC-3 ACM Codec and Dump");
        }

        void WriteWMA(IGraphBuilder pGraph, ICaptureGraphBuilder2 pBuilder, string outfile, IBaseFilter pSampleGrabber)
        {
            int hr = 0;
            //add WM ASF Writer
            IBaseFilter pWMASFWriter = (IBaseFilter)new WMAsfWriter();
            hr = pGraph.AddFilter(pWMASFWriter, "WM ASF Writer");
            checkHR(hr, "Can't add WM ASF Writer to graph");
            //set destination filename
            IFileSinkFilter pWMASFWriter_sink = pWMASFWriter as IFileSinkFilter;
            if (pWMASFWriter_sink == null)
                checkHR(unchecked((int)0x80004002), "Can't get IFileSinkFilter");
            hr = pWMASFWriter_sink.SetFileName(outfile, null);
            checkHR(hr, "Can't set filename");

            IConfigAsfWriter conf = (IConfigAsfWriter)pWMASFWriter;
            conf.ConfigureFilterUsingProfileGuid(new Guid("{407B9450-8BDC-4ee5-88B8-6F527BD941F2}")); //WMProfile_V80_128StereoAudio

            //connect SampleGrabber and WM ASF Writer
            hr = pBuilder.RenderStream(null, MediaType.Audio, pSampleGrabber, null, pWMASFWriter);
            checkHR(hr, "Can't connect SampleGrabber and WM ASF Writer");
        }



        public MainForm()
        {
            InitializeComponent();

            formatCombo.Items.Add(new OutputFormat("WAV", WriteWAV));
            formatCombo.Items.Add(new OutputFormat("MP3", WriteMP3));
            formatCombo.Items.Add(new OutputFormat("OGG", WriteOGG));
            formatCombo.Items.Add(new OutputFormat("AC3", WriteAC3));
            formatCombo.Items.Add(new OutputFormat("WMA", WriteWMA));
            formatCombo.SelectedIndex = 0;
        }

        private void OnBrowse(object sender, EventArgs e)
        {
            OpenFileDialog fd = new OpenFileDialog();
            fd.Filter = "WAV, MP3, OGG, M4A, WMA, AC3 files|*.wav;*.mp3;*.ogg;*.m4a;*.wma;*.ac3|All files (*.*)|*.*";
            if (fd.ShowDialog() != DialogResult.OK)
                return;
            inputFileBox.Text = fd.FileName;
            MakeOutputFileName();
        }

        void MakeOutputFileName()
        {
            string fname = inputFileBox.Text;
            int i = fname.LastIndexOf('.');
                outputFileBox.Text = (i > 0? fname.Substring(0, i): fname)
                   + "-new." + formatCombo.SelectedItem.ToString().ToLowerInvariant();
        }

        private void OnFormatChanged(object sender, EventArgs e)
        {
            MakeOutputFileName();
        }

        static void checkHR(int hr, string msg)
        {
            if (hr < 0)
            {
                MessageBox.Show(msg, "error");
                DsError.ThrowExceptionForHR(hr);
            }
        }

        private void OnStart(object sender, EventArgs e)
        {
            try
            {
                btnStart.Enabled = false;
                Application.UseWaitCursor = true;
                Cursor.Current = Cursors.WaitCursor;

                IGraphBuilder graph = (IGraphBuilder)new FilterGraph();
                BuildGraph(graph, inputFileBox.Text, outputFileBox.Text, ((OutputFormat)formatCombo.SelectedItem).makeGraphProc);
                IMediaControl mediaControl = (IMediaControl)graph;
                IMediaEvent mediaEvent = (IMediaEvent)graph;
                int hr = mediaControl.Run();
                checkHR(hr, "Can't run the graph");

                EventCode evcode;
                hr = mediaEvent.WaitForCompletion(-1, out evcode);
                checkHR(hr, "error while processing");
                mediaControl.Stop();

                //clear the graph
                IEnumFilters enumfilters;
                hr = graph.EnumFilters(out enumfilters);
                checkHR(hr, "can't enumerate filters");
                IBaseFilter[] flt = new IBaseFilter[1];
                IntPtr fetched = Marshal.AllocCoTaskMem(4);
                List<IBaseFilter> filters = new List<IBaseFilter>();
                while (enumfilters.Next(1, flt, fetched) == 0)
                {
                    if (Marshal.ReadInt32(fetched) < 1) break;
                    filters.Add(flt[0]);
                }
                foreach (IBaseFilter f in filters) {
                    graph.RemoveFilter(f);
                    Marshal.ReleaseComObject(f);
                }
                Marshal.ReleaseComObject(graph);
                Marshal.FreeCoTaskMem(fetched);
            }            
            catch (COMException ex)
            {
                MessageBox.Show(ex.Message, "COM Exception");
                return;
            }
            catch(Exception ex)
            {
                MessageBox.Show(ex.Message, "Exception");
                return;
            }
            finally
            {
                Cursor.Current = Cursors.Default;
                Application.UseWaitCursor = false;
                btnStart.Enabled = true;
                GC.Collect();
            }            
            MessageBox.Show("done", "Sound Converter");
        }


        void BuildGraph(IGraphBuilder pGraph, string infile, string outfile, WritingGraphPart MakeWritingPart)
        {
            IBaseFilter sourceFilter;
            int hr = pGraph.AddSourceFilter(infile, infile, out sourceFilter);
            checkHR(hr, "can't add source filter for " + infile);

            //for OGG splitting use specific filter
            if (infile.ToLowerInvariant().EndsWith(".ogg"))
            {
                Guid CLSID_HaaliMediaSplitterAR = new Guid("{564FD788-86C9-4444-971E-CC4A243DA150}"); //splitter.ax
                IBaseFilter pHaaliMediaSplitterAR = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_HaaliMediaSplitterAR));
                hr = pGraph.AddFilter(pHaaliMediaSplitterAR, "Haali Media Splitter (AR)");
                checkHR(hr, "Can't add Haali Media Splitter (AR) to graph");
            }

            //add SampleGrabber
            Guid CLSID_SampleGrabber = new Guid("{C1F400A0-3F08-11D3-9F0B-006008039E37}"); //qedit.dll
            IBaseFilter pSampleGrabber = (IBaseFilter)Activator.CreateInstance(Type.GetTypeFromCLSID(CLSID_SampleGrabber));
            hr = pGraph.AddFilter(pSampleGrabber, "SampleGrabber");
            checkHR(hr, "Can't add SampleGrabber to graph");

            //make sample grabber accept only PCM
            AMMediaType pSampleGrabber_pmt = new AMMediaType();
            pSampleGrabber_pmt.majorType = MediaType.Audio;
            pSampleGrabber_pmt.subType = MediaSubType.PCM;
            hr = ((ISampleGrabber)pSampleGrabber).SetMediaType(pSampleGrabber_pmt);
            DsUtils.FreeAMMediaType(pSampleGrabber_pmt);
            checkHR(hr, "Can't set media type to sample grabber");

            //graph builder helper
            ICaptureGraphBuilder2 pBuilder = (ICaptureGraphBuilder2)new CaptureGraphBuilder2();
            hr = pBuilder.SetFiltergraph(pGraph);
            checkHR(hr, "Can't SetFiltergraph");

            //connect source filter and SampleGrabber via splitter & decoder (if necessary)
            hr = pBuilder.RenderStream(null, null, sourceFilter, null, pSampleGrabber);
            checkHR(hr, "Can't connect source filter and SampleGrabber");

            MakeWritingPart(pGraph, pBuilder, outfile, pSampleGrabber);
        }

        static IPin GetPin(IBaseFilter filter, string pinname)
        {
            IEnumPins epins;
            int hr = filter.EnumPins(out epins);
            checkHR(hr, "Can't enumerate pins");
            IntPtr fetched = Marshal.AllocCoTaskMem(4);
            IPin[] pins = new IPin[1];
            while (epins.Next(1, pins, fetched) == 0)
            {
                PinInfo pinfo;
                pins[0].QueryPinInfo(out pinfo);
                bool found = (pinfo.name == pinname);
                DsUtils.FreePinInfo(pinfo);
                if (found)
                    return pins[0];
            }
            checkHR(-1, "Pin not found");
            return null;
        }

        public static IBaseFilter CreateFilter(string displayName)
        {
            int hr = 0;
            IBaseFilter filter = null;
            IBindCtx bindCtx = null;
            IMoniker moniker = null;

            try
            {
                hr = CreateBindCtx(0, out bindCtx);
                Marshal.ThrowExceptionForHR(hr);

                int eaten;
                hr = MkParseDisplayName(bindCtx, displayName, out eaten, out moniker);
                Marshal.ThrowExceptionForHR(hr);

                Guid guid = typeof(IBaseFilter).GUID;
                object obj;
                moniker.BindToObject(bindCtx, null, ref guid, out obj);
                filter = (IBaseFilter)obj;
            }
            finally
            {
                if (bindCtx != null) Marshal.ReleaseComObject(bindCtx);
                if (moniker != null) Marshal.ReleaseComObject(moniker);
            }

            return filter;
        }

        [DllImport("ole32.dll")]
        public static extern int CreateBindCtx(int reserved, out IBindCtx ppbc);

        [DllImport("ole32.dll")]
        public static extern int MkParseDisplayName(IBindCtx pcb, [MarshalAs(UnmanagedType.LPWStr)] string szUserName, out int pchEaten, out IMoniker ppmk);

    }
}